
hp_singlebases = [
    "shear",
    "stagger",
    "stretch",
    "buckle",
    "opening",
    "propel",
    "inclin",
    "tip",
    "xdisp",
    "ydisp"
]

hp_basepairs = [
    "majw",
    "majd",
    "minw",
    "mind",
    "rise",
    "shift",
    "slide",
    "roll",
    "tilt",
    "twist"
]

hp_angular = [
    "roll",
    "tilt",
    "twist",
    "buckle",
    "opening",
    "propel",
    "inclin",
    "tip"
]

hp_translational = [
    "shear",
    "stagger",
    "stretch",
    "xdisp",
    "ydisp",
    "majw",
    "majd",
    "minw",
    "mind",
    "rise",
    "shift",
    "slide"
]

hp_backbone = [
    "alphaC",
    "alphaW",
    "betaC",
    "betaW",
    "gammaC",
    "gammaW",
    "deltaC",
    "deltaW",
    "chiC",
    "chiW",
    "phaseC",
    "phaseW",
    "epsilC",
    "epsilW",
    "zetaC",
    "zetaW"
]

helical_parameters = hp_basepairs + hp_singlebases + hp_backbone
